from __future__ import absolute_import

from functools import partial

from graphene import relay
from graphene.relay.connection import PageInfo
from graphql_relay.connection.connectiontypes import Edge


class PynamoConnectionField(relay.ConnectionField):
    def __init__(self, type, *args, **kwargs):
        super(PynamoConnectionField, self).__init__(
            type,
            *args,
            **kwargs
        )

    @property
    def model(self):
        return self.type._meta.node._meta.model

    @classmethod
    def get_query(cls, model, context, info, args):
        # TODO this is very very basic and needs to handle connection params from args like first, after, etc.
        return model.scan()

    @classmethod
    def connection_resolver(cls, resolver, connection, model, root, args, context, info, **kwargs):
        iterable = resolver(root, args, context, info)
        if iterable is None and len(args):
            iterable = getattr(root, args[0])
        if iterable is None:
            iterable = cls.get_query(model, context, info, args)
        iterable = iter(iterable if iterable is not None else [])
        connection_type = connection
        edge_type = connection.Edge or Edge
        pageinfo_type = PageInfo

        full_args = dict(args, **kwargs)
        first = full_args.get('first')
        after = full_args.get('after')
        has_previous_page = bool(after)
        page_size = first if first else None

        start_cursor = None
        if after:
            for item in iterable:
                if item == after:
                    start_cursor = iterable.next()
                    break

        edges = []
        if page_size:
            while len(edges) < page_size:
                try:
                    entity = iterable.next()
                except StopIteration:
                    break

                edge = edge_type(node=entity, cursor=entity)
                edges.append(edge)
        else:
            edges = [edge_type(node=entity, cursor=entity) for entity in iterable]

        try:
            start_cursor = edges[0].node
            end_cursor = edges[-1].node
        except IndexError:
            end_cursor = None

        try:
            next(iterable)
            has_next = True
        except StopIteration:
            has_next = False

        # Construct the connection
        return connection_type(
            edges=edges,
            page_info=pageinfo_type(
                start_cursor=start_cursor if start_cursor else '',
                end_cursor=end_cursor if end_cursor else '',
                has_previous_page=has_previous_page,
                has_next_page=has_next
            )
        )

    def get_resolver(self, parent_resolver):
        return partial(self.connection_resolver, parent_resolver, self.type, self.model)



import logging

import graphene
from graphene import Node

from graphene_pynamodb.relationships import OneToOne, OneToMany
from .models import Reporter, Article
from ..types import PynamoObjectType

logging.basicConfig()


def setup_fixtures():
    class ReporterType(PynamoObjectType):
        class Meta:
            model = Reporter
            interfaces = (Node,)

    class ArticleType(PynamoObjectType):
        class Meta:
            model = Article
            interfaces = (Node,)

    reporter1 = Reporter(1, first_name="John", last_name="Snow")
    article1 = Article(1, headline="Awesome Article", reporter=reporter1)
    article2 = Article(2, headline="Lame Article", reporter=reporter1)
    reporter1.articles = [article1, article2]

    return {
        'ReporterType': ReporterType,
        'ArticleType': ArticleType,
        'reporter1': reporter1,
        'article1': article1,
        'article2': article2
    }


def test_relationships_should_resolve_well():
    fixtures = setup_fixtures()

    class Query(graphene.ObjectType):
        reporter = graphene.Field(fixtures["ReporterType"])
        articles = graphene.List(fixtures["ArticleType"])

        def resolve_reporter(self, *args, **kwargs):
            return fixtures["reporter1"]

        def resolve_articles(self, *args, **kwargs):
            return fixtures["reporter1"].articles

    query = '''
      query AuthorQuery {
        reporter {
          firstName,
          lastName,
          articles {
            edges {
              node {
                id
              }
            }
          }
        }
        articles {
          headline
          reporter {
            id
          }
        }
      }
    '''
    expected = {
        'reporter': {
            'firstName': 'John',
            'lastName': 'Snow',
            'articles': {
                'edges': [{
                    'node': {
                        'id': 'QXJ0aWNsZVR5cGU6MQ==',
                    }
                }, {
                    'node': {
                        'id': 'QXJ0aWNsZVR5cGU6Mg=='
                    }
                }]
            }
        },
        'articles': [{
            'headline': 'Awesome Article',
            'reporter': {
                'id': 'UmVwb3J0ZXJUeXBlOjE='
            }
        }, {
            'headline': 'Lame Article',
            'reporter': {
                'id': 'UmVwb3J0ZXJUeXBlOjE='
            }
        }]
    }

    schema = graphene.Schema(query=Query)

    result = schema.execute(query)
    assert not result.errors
    assert result.data['reporter'] == expected['reporter']
    assert all(item in result.data['articles'] for item in expected['articles'])


def test_onetoone_should_serialize_well():
    fixtures = setup_fixtures()
    relationship = OneToOne(Article)
    assert relationship.serialize(fixtures["article1"]) == "1"


def test_onetoone_should_deserialize_well():
    fixtures = setup_fixtures()
    relationship = OneToOne(Article)
    article = relationship.deserialize(2)
    assert isinstance(article, Article.__class__)
    assert article.id == fixtures["article2"].id


def test_onetomany_should_serialize_well():
    fixtures = setup_fixtures()
    relationship = OneToMany(Article)
    assert relationship.serialize([fixtures["article1"], fixtures["article2"]]) == ['1', '2']


def test_onetomany_should_deserialize_well():
    relationship = OneToMany(Article)

    articles = relationship.deserialize([1, 3])
    print(articles)

    assert len(articles) == 2
    assert isinstance(articles[0], Article.__class__)
    assert isinstance(articles[1], Article.__class__)
    assert articles[0].id == 1
    assert articles[1].id == 3
    assert articles[0].headline == "Hi!"
    assert articles[1].headline == "My Article"



from .fields import (
    PynamoConnectionField
)
from .types import (
    PynamoObjectType,
)

__all__ = ['PynamoObjectType',
           'PynamoConnectionField',
           'get_query',
           'get_session']
